// SortCheckListBox.cpp : implementation file
// Copyright (c) 2016 by David Wincelberg
//
// Window styles:
// Required: owner-draw fixed, has strings
// Recommended: horizontal scroll, vertical scroll, disable no scroll
// Optional: sort
// Prohibited: owner-draw variable
//
// Includes horizontal scrolling (based on the following reference).
//
// Reference:
// "How to implement horizontal scrolling in a CListBox-derived class" at:
// https://support.microsoft.com/en-us/kb/146437
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "SortCheckListBox.h"

#include "ListBoxData.h"
#include "ListBoxOrder.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CSortCheckListBox

// Note:	These parameters relate to the Strategy and Flyweight patterns.
CSortCheckListBox::CSortCheckListBox (CListBoxOrder* pListBoxOrder /* = nullptr */,
									  bool bHandleDelete /* = true */)
{
	m_pListBoxOrder  = (pListBoxOrder == nullptr ? new CListBoxOrder() : pListBoxOrder);
	m_bHandleDelete  = bHandleDelete;
	m_bSorting       = false;
	m_bSorted        = true;
	m_bAllowOrderMod = true;

	// Stores pointer for horizontal-extent calculations.
	SetBoxPointer (this, 11);	// offset for check boxes
}	// CSortCheckListBox


// Note:	These parameters relate to the Strategy and Flyweight patterns.
CSortCheckListBox::CSortCheckListBox (CStringOrder* pStringOrder,
									  bool bHandleDeleteInner /* = true */,
									  bool bHandleDeleteOuter /* = true */)
{
	m_pListBoxOrder  = new CListBoxOrder (pStringOrder, bHandleDeleteInner);
	m_bHandleDelete  = bHandleDeleteOuter;
	m_bSorting       = false;
	m_bSorted        = true;
	m_bAllowOrderMod = true;

	// Stores pointer for horizontal-extent calculations.
	SetBoxPointer (this, 11);	// offset for check boxes
}	// CSortCheckListBox


CSortCheckListBox::~CSortCheckListBox()
{
	if (m_bHandleDelete)
		delete m_pListBoxOrder;
}	// ~CSortCheckListBox


BEGIN_MESSAGE_MAP(CSortCheckListBox, CCheckListBox)
	//{{AFX_MSG_MAP(CSortCheckListBox)
	//}}AFX_MSG_MAP
	ON_MESSAGE (LB_SETTABSTOPS, OnSetTabStops)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSortCheckListBox message handlers

// Action:	Checks for required styles and sets the sorting flag.
void CSortCheckListBox::PreSubclassWindow()
{
	DWORD dwStyle = GetStyle();
	if ((dwStyle & LBS_OWNERDRAWFIXED) == 0 || (dwStyle & LBS_HASSTRINGS) == 0 ||
	(dwStyle & LBS_OWNERDRAWVARIABLE)) {
		TRACE(_T("** For CSortCheckListBox, set styles owner-draw fixed and has strings.\n"));
		TRACE(_T("** Do not set style owner-draw variable.\n"));
		ASSERT(false);
	}

	m_bSorting = ((dwStyle & LBS_SORT) != 0);

	CCheckListBox::PreSubclassWindow();
}	// PreSubclassWindow


// Action:	Replaces the sort-order object.
//			Deletes the old one unless m_bHandleDelete is false.
// Note:	This relates to the Strategy and Flyweight patterns.
void CSortCheckListBox::SetSortOrder (CListBoxOrder* pListBoxOrder,
									  bool bHandleDelete /* = true */)
{
	if (!m_bAllowOrderMod) {
		TRACE (_T("** Changing the sort-order strategy is not allowed ")
			_T("after adding or inserting a string.\n"));
		ASSERT (false);
		return;
	}

	if (pListBoxOrder == nullptr)
		return;

	if (m_bHandleDelete)
		delete m_pListBoxOrder;

	m_pListBoxOrder = pListBoxOrder;
	m_bHandleDelete = bHandleDelete;
}	// SetSortOrder


// Action:	Replaces the string-order object.
//			Deletes the old one unless CListBoxOrder's m_bHandleDelete is false.
// Note:	This relates to the Strategy and Flyweight patterns.
void CSortCheckListBox::SetStringOrder (CStringOrder* pStringOrder,
										bool bHandleDelete /*  = true */)
{
	if (!m_bAllowOrderMod) {
		TRACE (_T("** Changing the string-order strategy is not allowed ")
			_T("after adding or inserting a string.\n"));
		ASSERT (false);
		return;
	}

	if (pStringOrder == nullptr || m_pListBoxOrder == nullptr)
		return;

	m_pListBoxOrder->SetStringOrder (pStringOrder, bHandleDelete);
}	// SetStringOrder


// Action:	Checks if a list-box-data object was supplied; o.w. allocates one.
bool CSortCheckListBox::CheckListBoxData (CListBoxData** ppListBoxData)
{
	if (*ppListBoxData == nullptr) {
		try {
			*ppListBoxData = new CListBoxData();
		}
		catch (CMemoryException* e) {
			e->Delete();
			return false;
		}
	}

	return true;
}	// CheckListBoxData


// Action:	Prepare a string by handling a check code and adding a space (if specified)
//			so that the text is not too close to the check box.
// Throws:	CMemoryException: Delete, +
void CSortCheckListBox::PrepareString (CString& strItem, int* pnCheck,
									   bool bHasCheckCode, bool bAddSpace)
{
	if (strItem.IsEmpty())
		return;

	if (bHasCheckCode) {
		if (::isdigit (strItem[0])) {
			_TCHAR szCheck[2] = {strItem[0], _T('\0')};
			int n = ::_ttoi (szCheck);	// generic for atoi
			if (n < 3) {	// Valid check states are 0, 1 and 2.
				*pnCheck = n;
				if (bAddSpace)
					strItem.SetAt (0, ' ');
				else
					strItem.Delete (0);
			}
		}
	}
	else if (bAddSpace)
		strItem = _T(" ") + strItem;
} // PrepareString


// Action:	Adds a string to the list box and stores extent data.
//			If sorting is selected, sorts based on the current string order.
//			The first character may contain a check code.
// Note:	Allows for storing a data item that is in a subclass of CListBoxData.
int CSortCheckListBox::AddString (LPCTSTR lpszItem, bool bAddSpace /* = true */,
								  int nCheckDefault /* = 0 */,
								  bool bHasCheckCode /* = false */,
								  CListBoxData* pListBoxData /* = nullptr */)
{
	if (lpszItem == nullptr) {
		delete pListBoxData;
		return LB_ERR;
	}

	if (!CheckListBoxData (&pListBoxData))
		return LB_ERR;

	CString strItem;
	int nCheck = nCheckDefault;

	try {
		strItem = lpszItem;
		PrepareString (strItem, &nCheck, bHasCheckCode, bAddSpace);
	}
	catch (CMemoryException* e) { // =, PrepareString
		e->Delete();
		return LB_ERR;
	}

	if (!m_bSorting) {
		int nResult = CListBox::AddString (strItem);
		if (nResult == LB_ERR || nResult == LB_ERRSPACE) {
			delete pListBoxData;
			return nResult;
		}

		StoreExtent (nResult, strItem, pListBoxData);
		SetCheck (nResult, nCheck);
		m_bAllowOrderMod = false;
		return nResult;
	}

	int nPosition = 0, nCount = GetCount();

	if (nCount > 0 && m_pListBoxOrder != nullptr) {
		try {
			nPosition = SortStrings (strItem, pListBoxData, nCount);
		}
		catch (CMemoryException* e) {
			e->Delete();
			return LB_ERR;
		}
	}

	int nResult = CListBox::InsertString (nPosition, strItem);
	if (nResult == LB_ERR || nResult == LB_ERRSPACE) {
		delete pListBoxData;
		return nResult;
	}

	StoreExtent (nResult, strItem, pListBoxData);
	SetCheck (nResult, nCheck);
	m_bAllowOrderMod = false;
	return nResult;
}	// AddString


// Action:	Adds a string to a sorted list or one with insertions.
// Throws:	CMemoryException: =, GetText
int CSortCheckListBox::SortStrings (CString& strItem1, CListBoxData* pListBoxData1,
									int nCount)
{
	CString strItem2;
	CListBoxData* pListBoxData2;

	if (m_bSorted) {
		// Binary search insertion
		int nLower = 0, nUpper = nCount - 1, nMid, nCompare;

		while (nLower <= nUpper) {
			nMid = (nLower + nUpper) / 2;
			GetText (nMid, strItem2);
			pListBoxData2 = GetItemDataObjPtr (nMid);

			nCompare = m_pListBoxOrder->Compare (strItem1, pListBoxData1,
				strItem2, pListBoxData2);

			if (nCompare > 0)
				nLower = nMid + 1;
			else if (nCompare < 0)
				nUpper = nMid - 1;
			else
				break;
		}

		// When a match is not found:
		// nLower <= position <= nUpper + 1 == nLower
		// See http://www.FileJockeySoftware.com/ANO/BinSearchInsert.html.
		return (nCompare == 0 ? nMid : nLower);
	}
	else {
		// Linear search insertion: only considers added strings
		for (int n = 0;  n < nCount;  ++n) {
			pListBoxData2 = GetItemDataObjPtr (n);

			if (pListBoxData2 == nullptr)
				return -1;
			else if (pListBoxData2->WasAdded()) {
				GetText (n, strItem2);

				if (m_pListBoxOrder->Compare (strItem1, pListBoxData1,
				strItem2, pListBoxData2) < 0)
					return n;
			}
		}

		return -1;	// after last string
	}
}	// SortStrings


// Action:	Inserts a string into a check list box at a specified position.
// Note:	Allows for storing a data item that is in a subclass of CListBoxData.
int CSortCheckListBox::InsertString (int nIndex, LPCTSTR lpszItem,
									 bool bAddSpace /* = true */,
									 int nCheckDefault /* = 0 */,
									 bool bHasCheckCode /* = false */,
									 CListBoxData* pListBoxData /* = nullptr */)
{
	if (lpszItem == nullptr) {
		delete pListBoxData;
		return LB_ERR;
	}

	if (!CheckListBoxData (&pListBoxData))
		return LB_ERR;

	CString strItem;
	int nCheck = nCheckDefault;

	try {
		strItem = lpszItem;
		PrepareString (strItem, &nCheck, bHasCheckCode, bAddSpace);
	}
	catch (CMemoryException* e) { // =, PrepareString
		e->Delete();
		return LB_ERR;
	}

	int nResult = CListBox::InsertString (nIndex, strItem);
	if (nResult == LB_ERR || nResult == LB_ERRSPACE) {
		delete pListBoxData;
		return nResult;
	}

	pListBoxData->SetInserted();
	StoreExtent (nResult, strItem, pListBoxData);
	SetCheck (nResult, nCheck);
	m_bSorted        = false;
	m_bAllowOrderMod = false;
	return nResult;
}	// InsertString


// Action:	Deletes a string, may adjust the horizontal extent and
//			may allow the sort and string orders to change.
int CSortCheckListBox::DeleteString (UINT nIndex)
{
	CListBoxData* pListBoxData = GetItemDataObjPtr (nIndex);
	int nResult;

	if (pListBoxData != nullptr) {
		int nExtent = pListBoxData->GetExtent();
		nResult = CListBox::DeleteString (nIndex);
		if (nResult == LB_ERR)
			return LB_ERR;

		bool bRecalc = (nExtent >= m_nMaxExtent);

		if (m_bDefer)
			m_bNeedRecalc |= bRecalc;
		else if (bRecalc)
			UpdateMaxExtent();
	}
	else
		nResult = CListBox::DeleteString (nIndex);

	if (GetCount() == 0)
		ResetContent();

	return nResult;
}	// DeleteString


// Action:	Releases memory that is no longer needed.
// Note:	The subclass destructor will be called.
void CSortCheckListBox::DeleteItem (LPDELETEITEMSTRUCT lpDeleteItemStruct)
{
	delete reinterpret_cast<CListBoxData*> (lpDeleteItemStruct->itemData);
}	// DeleteItem


// Action:	Clears the list box, resets the horizontal extent
//			and allows for the sort and string orders to be changed.
void CSortCheckListBox::ResetContent()
{
	CListBox::ResetContent();
	SetHorizontalExtent (0);
	m_nMaxExtent     = 0;
	m_bDefer         = false;
	m_bNeedRecalc    = false;
	m_bSorted        = true;
	m_bAllowOrderMod = true;
}	// ResetContent


// Action:	Stores user data in a CListBoxData object.
int CSortCheckListBox::SetItemData (int nIndex, DWORD dwItemData)
{
	CListBoxData* pListBoxData = GetItemDataObjPtr (nIndex);

	if (pListBoxData != nullptr) {
		pListBoxData->SetItemData (dwItemData);
		return TRUE;
	}
	else
		return LB_ERR;
}	// SetItemData


// Action:	Retrieves user data from a CListBoxData object.
DWORD CSortCheckListBox::GetItemData (int nIndex) const
{
	CListBoxData* pListBoxData = GetItemDataObjPtr (nIndex);

	if (pListBoxData != nullptr)
		return pListBoxData->GetItemData();
	else
		return LB_ERR;
}	// GetItemData


// Action:	Stores user data in a CListBoxData object.
int CSortCheckListBox::SetItemDataPtr (int nIndex, void* pData)
{
	CListBoxData* pListBoxData = GetItemDataObjPtr (nIndex);

	if (pListBoxData != nullptr) {
		pListBoxData->SetItemData (reinterpret_cast<DWORD>(pData));
		return TRUE;
	}
	else
		return LB_ERR;
}	// SetItemDataPtr


// Action:	Retrieves user data from a CListBoxData object.
void* CSortCheckListBox::GetItemDataPtr (int nIndex) const
{
	CListBoxData* pListBoxData = GetItemDataObjPtr (nIndex);

	if (pListBoxData != nullptr)
		return reinterpret_cast<void*> (pListBoxData->GetItemData());
	else
		return reinterpret_cast<void*> (LB_ERR);
}	// GetItemDataPtr


// Action:	Stores user data as a pointer to a CListBoxData object.
// Note:	Copies the data values from the old data object and deletes it.
//			pListBoxData should be from a subclass of CListBoxData.
int CSortCheckListBox::SetItemDataObjPtr (int nIndex, CListBoxData* pListBoxData)
{
	if (pListBoxData == nullptr)
		return LB_ERR;

	CListBoxData* pOldListBoxData = GetItemDataObjPtr (nIndex);

	if (pOldListBoxData != nullptr) {
		pListBoxData = pOldListBoxData;
		delete pOldListBoxData;
		return CListBox::SetItemData (nIndex, reinterpret_cast<DWORD> (pListBoxData));
	}
	else
		return LB_ERR;
}	// SetItemDataObjPtr


// Action:	Retrieves user data as a pointer to a CListBoxData object.
CListBoxData* CSortCheckListBox::GetItemDataObjPtr (int nIndex) const
{
	int nCount = GetCount();

	if (nCount == LB_ERR || nIndex < 0 || nIndex >= nCount)
		return nullptr;

	DWORD dwData = CListBox::GetItemData (nIndex);

	if (dwData != LB_ERR)
		return reinterpret_cast<CListBoxData*> (dwData);
	else
		return nullptr;
}	// GetItemDataObjPtr

/////////////////////////////////////////////////////////////////////////////
// Extent function

// Action:	Intercepts the LB_SETTABSTOPS message to update the tab stop array.
LRESULT CSortCheckListBox::OnSetTabStops (WPARAM wParam, LPARAM lParam)
{
	LRESULT lResult = Default();
	if (!lResult)
		return lResult;

	return TabStopsSet (wParam, lParam, lResult);
}	// OnSetTabStops
